package utils;

import java.math.*;

public final class BigSquareRoot {
	
	private static BigDecimal TWO = new BigDecimal(2);
	
	public static final int MAX_ITERATIONS = 50;

	private static BigDecimal getInitialApproximation(BigDecimal n) {
		BigInteger integerPart = n.toBigInteger();
		int length = integerPart.toString().length();
		if ((length % 2) == 0) {
			length--;
		}
		length /= 2;
		BigDecimal guess = BigDecimal.ONE.movePointRight(length);
		return guess;
	}

	public static BigDecimal sqrt(BigDecimal x, MathContext context) {
		if (x.compareTo(BigDecimal.ZERO) <= 0) {
			throw new IllegalArgumentException(x.toString());
		}

		BigDecimal initialGuess = getInitialApproximation(x);
		BigDecimal guess = new BigDecimal(initialGuess.toString());
		BigDecimal error;

		// Iterate

		int iterations = 0;
		boolean more = true;
		while (more) {
			BigDecimal lastGuess = guess;
			guess = x.divide(guess, context);
			guess = guess.add(lastGuess, context);
			guess = guess.divide(TWO, context);
			error = x.subtract(guess.multiply(guess));
			if (++iterations >= MAX_ITERATIONS) {
				more = false;
			} else if (lastGuess.equals(guess)) {
				more = error.abs().compareTo(BigDecimal.ONE) >= 0;
			}
		}

		return guess;
	}

}
